@@ -22,25 +22,25 @@ use std::sync::Arc;
 
 use base64::prelude::BASE64_STANDARD;
 use base64::Engine;
-use bytes::Buf;
-use http::header::CONTENT_TYPE;
 use http::Response;
 use http::StatusCode;
 use log::debug;
 use reqsign::AzureStorageConfig;
 use reqsign::AzureStorageLoader;
 use reqsign::AzureStorageSigner;
-use serde::Deserialize;
-use serde::Serialize;
 use sha2::Digest;
 use sha2::Sha256;
 
+use super::core::constants::X_MS_META_PREFIX;
+use super::core::constants::X_MS_VERSION_ID;
+use super::core::AzblobCore;
+use super::delete::AzblobDeleter;
 use super::error::parse_error;
 use super::lister::AzblobLister;
 use super::writer::AzblobWriter;
+use super::writer::AzblobWriters;
 use crate::raw::*;
-use crate::services::azblob::core::AzblobCore;
-use crate::services::azblob::writer::AzblobWriters;
+use crate::services::AzblobConfig;
 use crate::*;
 
 /// Known endpoint suffix Azure Storage Blob services resource URI syntax.
@@ -55,74 +55,14 @@ const KNOWN_AZBLOB_ENDPOINT_SUFFIX: &[&str] = &[
 
 const AZBLOB_BATCH_LIMIT: usize = 256;
 
-/// Azure Storage Blob services support.
-#[derive(Default, Serialize, Deserialize, Clone, PartialEq, Eq)]
-pub struct AzblobConfig {
-    /// The root of Azblob service backend.
-    ///
-    /// All operations will happen under this root.
-    pub root: Option<String>,
-
-    /// The container name of Azblob service backend.
-    pub container: String,
-
-    /// The endpoint of Azblob service backend.
-    ///
-    /// Endpoint must be full uri, e.g.
-    ///
-    /// - Azblob: `https://accountname.blob.core.windows.net`
-    /// - Azurite: `http://127.0.0.1:10000/devstoreaccount1`
-    pub endpoint: Option<String>,
-
-    /// The account name of Azblob service backend.
-    pub account_name: Option<String>,
-
-    /// The account key of Azblob service backend.
-    pub account_key: Option<String>,
-
-    /// The encryption key of Azblob service backend.
-    pub encryption_key: Option<String>,
-
-    /// The encryption key sha256 of Azblob service backend.
-    pub encryption_key_sha256: Option<String>,
-
-    /// The encryption algorithm of Azblob service backend.
-    pub encryption_algorithm: Option<String>,
-
-    /// The sas token of Azblob service backend.
-    pub sas_token: Option<String>,
-
-    /// The maximum batch operations of Azblob service backend.
-    pub batch_max_operations: Option<usize>,
-}
-
-impl Debug for AzblobConfig {
-    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
-        let mut ds = f.debug_struct("AzblobConfig");
-
-        ds.field("root", &self.root);
-        ds.field("container", &self.container);
-        ds.field("endpoint", &self.endpoint);
-
-        if self.account_name.is_some() {
-            ds.field("account_name", &"<redacted>");
-        }
-        if self.account_key.is_some() {
-            ds.field("account_key", &"<redacted>");
-        }
-        if self.sas_token.is_some() {
-            ds.field("sas_token", &"<redacted>");
-        }
-
-        ds.finish()
-    }
-}
-
 impl Configurator for AzblobConfig {
     type Builder = AzblobBuilder;
+
+    #[allow(deprecated)]
     fn into_builder(self) -> Self::Builder {
         AzblobBuilder {
             config: self,
+
             http_client: None,
         }
     }
@@ -132,6 +72,8 @@ impl Configurator for AzblobConfig {
 #[derive(Default, Clone)]
 pub struct AzblobBuilder {
     config: AzblobConfig,
+
+    #[deprecated(since = "0.53.0", note = "Use `Operator::update_http_client` instead")]
     http_client: Option<HttpClient>,
 }
 
@@ -150,9 +92,11 @@ impl AzblobBuilder {
     ///
     /// All operations will happen under this root.
     pub fn root(mut self, root: &str) -> Self {
-        if !root.is_empty() {
-            self.config.root = Some(root.to_string())
-        }
+        self.config.root = if root.is_empty() {
+            None
+        } else {
+            Some(root.to_string())
+        };
 
         self
     }
@@ -306,6 +250,8 @@ impl AzblobBuilder {
     ///
     /// This API is part of OpenDAL's Raw API. `HttpClient` could be changed
     /// during minor updates.
+    #[deprecated(since = "0.53.0", note = "Use `Operator::update_http_client` instead")]
+    #[allow(deprecated)]
     pub fn http_client(mut self, client: HttpClient) -> Self {
         self.http_client = Some(client);
         self
@@ -431,15 +377,6 @@ impl Builder for AzblobBuilder {
         }?;
         debug!("backend use endpoint {}", &container);
 
-        let client = if let Some(client) = self.http_client {
-            client
-        } else {
-            HttpClient::new().map_err(|err| {
-                err.with_operation("Builder::build")
-                    .with_context("service", Scheme::Azblob)
-            })?
-        };
-
         let mut config_loader = AzureStorageConfig::default().from_env();
 
         if let Some(v) = self
@@ -494,13 +431,76 @@ impl Builder for AzblobBuilder {
 
         let signer = AzureStorageSigner::new();
 
-        let batch_max_operations = self
-            .config
-            .batch_max_operations
-            .unwrap_or(AZBLOB_BATCH_LIMIT);
-
         Ok(AzblobBackend {
             core: Arc::new(AzblobCore {
+                info: {
+                    let am = AccessorInfo::default();
+                    am.set_scheme(Scheme::Azblob)
+                        .set_root(&root)
+                        .set_name(container)
+                        .set_native_capability(Capability {
+                            stat: true,
+                            stat_with_if_match: true,
+                            stat_with_if_none_match: true,
+                            stat_has_cache_control: true,
+                            stat_has_content_length: true,
+                            stat_has_content_type: true,
+                            stat_has_content_encoding: true,
+                            stat_has_content_range: true,
+                            stat_has_etag: true,
+                            stat_has_content_md5: true,
+                            stat_has_last_modified: true,
+                            stat_has_content_disposition: true,
+
+                            read: true,
+
+                            read_with_if_match: true,
+                            read_with_if_none_match: true,
+                            read_with_override_content_disposition: true,
+                            read_with_if_modified_since: true,
+                            read_with_if_unmodified_since: true,
+
+                            write: true,
+                            write_can_append: true,
+                            write_can_empty: true,
+                            write_can_multi: true,
+                            write_with_cache_control: true,
+                            write_with_content_type: true,
+                            write_with_if_not_exists: true,
+                            write_with_if_none_match: true,
+                            write_with_user_metadata: true,
+
+                            delete: true,
+                            delete_max_size: Some(AZBLOB_BATCH_LIMIT),
+
+                            copy: true,
+
+                            list: true,
+                            list_with_recursive: true,
+                            list_has_etag: true,
+                            list_has_content_length: true,
+                            list_has_content_md5: true,
+                            list_has_content_type: true,
+                            list_has_last_modified: true,
+
+                            presign: self.config.sas_token.is_some(),
+                            presign_stat: self.config.sas_token.is_some(),
+                            presign_read: self.config.sas_token.is_some(),
+                            presign_write: self.config.sas_token.is_some(),
+
+                            shared: true,
+
+                            ..Default::default()
+                        });
+
+                    // allow deprecated api here for compatibility
+                    #[allow(deprecated)]
+                    if let Some(client) = self.http_client {
+                        am.update_http_client(|_| client);
+                    }
+
+                    am.into()
+                },
                 root,
                 endpoint,
                 encryption_key,
@@ -508,12 +508,9 @@ impl Builder for AzblobBuilder {
                 encryption_algorithm,
                 container: self.config.container.clone(),
 
-                client,
                 loader: cred_loader,
                 signer,
-                batch_max_operations,
             }),
-            has_sas_token: self.config.sas_token.is_some(),
         })
     }
 }
@@ -546,59 +543,20 @@ fn infer_storage_name_from_endpoint(endpoint: &str) -> Option<String> {
 #[derive(Debug, Clone)]
 pub struct AzblobBackend {
     core: Arc<AzblobCore>,
-    has_sas_token: bool,
 }
 
 impl Access for AzblobBackend {
     type Reader = HttpBody;
     type Writer = AzblobWriters;
     type Lister = oio::PageLister<AzblobLister>;
+    type Deleter = oio::BatchDeleter<AzblobDeleter>;
     type BlockingReader = ();
     type BlockingWriter = ();
     type BlockingLister = ();
+    type BlockingDeleter = ();
 
     fn info(&self) -> Arc<AccessorInfo> {
-        let mut am = AccessorInfo::default();
-        am.set_scheme(Scheme::Azblob)
-            .set_root(&self.core.root)
-            .set_name(&self.core.container)
-            .set_native_capability(Capability {
-                stat: true,
-                stat_with_if_match: true,
-                stat_with_if_none_match: true,
-
-                read: true,
-
-                read_with_if_match: true,
-                read_with_if_none_match: true,
-                read_with_override_content_disposition: true,
-
-                write: true,
-                write_can_append: true,
-                write_can_empty: true,
-                write_can_multi: true,
-                write_with_cache_control: true,
-                write_with_content_type: true,
-
-                delete: true,
-                copy: true,
-
-                list: true,
-                list_with_recursive: true,
-
-                presign: self.has_sas_token,
-                presign_stat: self.has_sas_token,
-                presign_read: self.has_sas_token,
-                presign_write: self.has_sas_token,
-
-                batch: true,
-                batch_delete: true,
-                batch_max_operations: Some(self.core.batch_max_operations),
-
-                ..Default::default()
-            });
-
-        am.into()
+        self.core.info.clone()
     }
 
     async fn stat(&self, path: &str, args: OpStat) -> Result<RpStat> {
@@ -607,8 +565,21 @@ impl Access for AzblobBackend {
         let status = resp.status();
 
         match status {
-            StatusCode::OK => parse_into_metadata(path, resp.headers()).map(RpStat::new),
-            _ => Err(parse_error(resp).await?),
+            StatusCode::OK => {
+                let headers = resp.headers();
+                let mut meta = parse_into_metadata(path, headers)?;
+                if let Some(version_id) = parse_header_to_str(headers, X_MS_VERSION_ID)? {
+                    meta.set_version(version_id);
+                }
+
+                let user_meta = parse_prefixed_headers(headers, X_MS_META_PREFIX);
+                if !user_meta.is_empty() {
+                    meta.with_user_metadata(user_meta);
+                }
+
+                Ok(RpStat::new(meta))
+            }
+            _ => Err(parse_error(resp)),
         }
     }
 
@@ -621,7 +592,7 @@ impl Access for AzblobBackend {
             _ => {
                 let (part, mut body) = resp.into_parts();
                 let buf = body.to_buffer().await?;
-                Err(parse_error(Response::from_parts(part, buf)).await?)
+                Err(parse_error(Response::from_parts(part, buf)))
             }
         }
     }
@@ -632,8 +603,8 @@ impl Access for AzblobBackend {
             AzblobWriters::Two(oio::AppendWriter::new(w))
         } else {
             AzblobWriters::One(oio::BlockWriter::new(
+                self.core.info.clone(),
                 w,
-                args.executor().cloned(),
                 args.concurrent(),
             ))
         };
@@ -641,15 +612,11 @@ impl Access for AzblobBackend {
         Ok((RpWrite::default(), w))
     }
 
-    async fn delete(&self, path: &str, _: OpDelete) -> Result<RpDelete> {
-        let resp = self.core.azblob_delete_blob(path).await?;
-
-        let status = resp.status();
-
-        match status {
-            StatusCode::ACCEPTED | StatusCode::NOT_FOUND => Ok(RpDelete::default()),
-            _ => Err(parse_error(resp).await?),
-        }
+    async fn delete(&self) -> Result<(RpDelete, Self::Deleter)> {
+        Ok((
+            RpDelete::default(),
+            oio::BatchDeleter::new(AzblobDeleter::new(self.core.clone())),
+        ))
     }
 
     async fn list(&self, path: &str, args: OpList) -> Result<(RpList, Self::Lister)> {
@@ -670,23 +637,29 @@ impl Access for AzblobBackend {
 
         match status {
             StatusCode::ACCEPTED => Ok(RpCopy::default()),
-            _ => Err(parse_error(resp).await?),
+            _ => Err(parse_error(resp)),
         }
     }
 
     async fn presign(&self, path: &str, args: OpPresign) -> Result<RpPresign> {
-        let mut req = match args.operation() {
-            PresignOperation::Stat(v) => self.core.azblob_head_blob_request(path, v)?,
+        let req = match args.operation() {
+            PresignOperation::Stat(v) => self.core.azblob_head_blob_request(path, v),
             PresignOperation::Read(v) => {
                 self.core
-                    .azblob_get_blob_request(path, BytesRange::default(), v)?
+                    .azblob_get_blob_request(path, BytesRange::default(), v)
             }
             PresignOperation::Write(_) => {
                 self.core
-                    .azblob_put_blob_request(path, None, &OpWrite::default(), Buffer::new())?
+                    .azblob_put_blob_request(path, None, &OpWrite::default(), Buffer::new())
             }
+            PresignOperation::Delete(_) => Err(Error::new(
+                ErrorKind::Unsupported,
+                "operation is not supported",
+            )),
         };
 
+        let mut req = req?;
+
         self.core.sign_query(&mut req).await?;
 
         let (parts, _) = req.into_parts();
@@ -697,77 +670,6 @@ impl Access for AzblobBackend {
             parts.headers,
         )))
     }
-
-    async fn batch(&self, args: OpBatch) -> Result<RpBatch> {
-        let ops = args.into_operation();
-        let paths = ops.into_iter().map(|(p, _)| p).collect::<Vec<_>>();
-        if paths.len() > AZBLOB_BATCH_LIMIT {
-            return Err(Error::new(
-                ErrorKind::Unsupported,
-                "batch delete limit exceeded",
-            ));
-        }
-
-        // construct and complete batch request
-        let resp = self.core.azblob_batch_delete(&paths).await?;
-
-        // check response status
-        if resp.status() != StatusCode::ACCEPTED {
-            return Err(parse_error(resp).await?);
-        }
-
-        // get boundary from response header
-        let content_type = resp.headers().get(CONTENT_TYPE).ok_or_else(|| {
-            Error::new(
-                ErrorKind::Unexpected,
-                "response data should have CONTENT_TYPE header",
-            )
-        })?;
-        let content_type = content_type
-            .to_str()
-            .map(|ty| ty.to_string())
-            .map_err(|e| {
-                Error::new(
-                    ErrorKind::Unexpected,
-                    format!("get invalid CONTENT_TYPE header in response: {:?}", e),
-                )
-            })?;
-        let splits = content_type.split("boundary=").collect::<Vec<&str>>();
-        let boundary = splits.get(1).to_owned().ok_or_else(|| {
-            Error::new(
-                ErrorKind::Unexpected,
-                "No boundary message provided in CONTENT_TYPE",
-            )
-        })?;
-
-        let mut bs = resp.into_body();
-        let bs = bs.copy_to_bytes(bs.remaining());
-
-        let multipart: Multipart<MixedPart> = Multipart::new().with_boundary(boundary).parse(bs)?;
-        let parts = multipart.into_parts();
-
-        if paths.len() != parts.len() {
-            return Err(Error::new(
-                ErrorKind::Unexpected,
-                "invalid batch response, paths and response parts don't match",
-            ));
-        }
-
-        let mut results = Vec::with_capacity(parts.len());
-
-        for (i, part) in parts.into_iter().enumerate() {
-            let resp = part.into_response();
-            let path = paths[i].clone();
-
-            // deleting not existing objects is ok
-            if resp.status() == StatusCode::ACCEPTED || resp.status() == StatusCode::NOT_FOUND {
-                results.push((path, Ok(RpDelete::default().into())));
-            } else {
-                results.push((path, Err(parse_error(resp).await?)));
-            }
-        }
-        Ok(RpBatch::new(results))
-    }
 }
 
 #[cfg(test)]
@@ -838,7 +740,7 @@ TableEndpoint=http://127.0.0.1:10002/devstoreaccount1;
 SharedAccessSignature=sv=2021-01-01&ss=b&srt=c&sp=rwdlaciytfx&se=2022-01-01T11:00:14Z&st=2022-01-02T03:00:14Z&spr=https&sig=KEllk4N8f7rJfLjQCmikL2fRVt%2B%2Bl73UBkbgH%2FK3VGE%3D
         "#,
         )
-        .expect("from connection string must succeed");
+            .expect("from connection string must succeed");
 
         assert_eq!(
             builder.config.endpoint.unwrap(),
@@ -859,7 +761,7 @@ AccountKey=account-key;
 SharedAccessSignature=sv=2021-01-01&ss=b&srt=c&sp=rwdlaciytfx&se=2022-01-01T11:00:14Z&st=2022-01-02T03:00:14Z&spr=https&sig=KEllk4N8f7rJfLjQCmikL2fRVt%2B%2Bl73UBkbgH%2FK3VGE%3D
         "#,
         )
-        .expect("from connection string must succeed");
+            .expect("from connection string must succeed");
 
         // SAS should be preferred over shared key
         assert_eq!(builder.config.sas_token.unwrap(), "sv=2021-01-01&ss=b&srt=c&sp=rwdlaciytfx&se=2022-01-01T11:00:14Z&st=2022-01-02T03:00:14Z&spr=https&sig=KEllk4N8f7rJfLjQCmikL2fRVt%2B%2Bl73UBkbgH%2FK3VGE%3D");
